@@ -31,6 +31,9 @@ WINE_DEFAULT_DEBUG_CHANNEL(dwrite);
 extern const unsigned short wine_linebreak_table[] DECLSPEC_HIDDEN;
 extern const unsigned short wine_scripts_table[] DECLSPEC_HIDDEN;
 
+/* Number of characters needed for LOCALE_SNATIVEDIGITS */
+#define NATIVE_DIGITS_LEN 11
+
 struct dwritescript_properties {
     DWRITE_SCRIPT_PROPERTIES props;
     UINT32 scripttag;    /* OpenType script tag */
@@ -186,7 +189,8 @@ static const struct dwritescript_properties dwritescripts_properties[Script_Last
 
 const char *debugstr_sa_script(UINT16 script)
 {
-    return script < Script_LastId ? debugstr_an((char*)&dwritescripts_properties[script].props.isoScriptCode, 4): "not defined";
+    return script < Script_LastId ? debugstr_an((char*)&dwritescripts_properties[script].props.isoScriptCode, 4):
+            "undefined";
 }
 
 /* system font falback configuration */
@@ -194,18 +198,18 @@ static const WCHAR meiryoW[] = {'M','e','i','r','y','o',0};
 
 struct fallback_mapping {
     DWRITE_UNICODE_RANGE range;
-    const WCHAR *family;
+    const WCHAR *families[5];
 };
 
 static const struct fallback_mapping fontfallback_neutral_data[] = {
-    { { 0x3000, 0x30ff }, meiryoW }, /* CJK Symbols and Punctuation, Hiragana, Katakana */
-    { { 0x31f0, 0x31ff }, meiryoW }, /* Katakana Phonetic Extensions */
-    { { 0x4e00, 0x9fff }, meiryoW }, /* CJK Unified Ideographs */
+    { { 0x3000, 0x30ff }, { meiryoW } }, /* CJK Symbols and Punctuation, Hiragana, Katakana */
+    { { 0x31f0, 0x31ff }, { meiryoW } }, /* Katakana Phonetic Extensions */
+    { { 0x4e00, 0x9fff }, { meiryoW } }, /* CJK Unified Ideographs */
 };
 
 struct dwrite_fontfallback {
     IDWriteFontFallback IDWriteFontFallback_iface;
-    IDWriteFactory4 *factory;
+    IDWriteFactory5 *factory;
     IDWriteFontCollection1 *systemcollection;
     const struct fallback_mapping *mappings;
     UINT32 count;
@@ -225,6 +229,8 @@ static inline struct dwrite_numbersubstitution *impl_from_IDWriteNumberSubstitut
     return CONTAINING_RECORD(iface, struct dwrite_numbersubstitution, IDWriteNumberSubstitution_iface);
 }
 
+static struct dwrite_numbersubstitution *unsafe_impl_from_IDWriteNumberSubstitution(IDWriteNumberSubstitution *iface);
+
 static inline struct dwrite_fontfallback *impl_from_IDWriteFontFallback(IDWriteFontFallback *iface)
 {
     return CONTAINING_RECORD(iface, struct dwrite_fontfallback, IDWriteFontFallback_iface);
@@ -363,20 +369,31 @@ enum linebreaking_classes {
     b_ZWJ,
 };
 
+static BOOL has_strong_condition(DWRITE_BREAK_CONDITION old_condition, DWRITE_BREAK_CONDITION new_condition)
+{
+    if (old_condition == DWRITE_BREAK_CONDITION_MAY_NOT_BREAK || old_condition == DWRITE_BREAK_CONDITION_MUST_BREAK)
+        return TRUE;
+
+    if (old_condition == DWRITE_BREAK_CONDITION_CAN_BREAK && new_condition != DWRITE_BREAK_CONDITION_MUST_BREAK)
+        return TRUE;
+
+    return FALSE;
+}
+
 /* "Can break" is a weak condition, stronger "may not break" and "must break" override it. Initially all conditions are
     set to "can break" and could only be changed once. */
 static inline void set_break_condition(UINT32 pos, enum BreakConditionLocation location, DWRITE_BREAK_CONDITION condition,
     struct linebreaking_state *state)
 {
     if (location == BreakConditionBefore) {
-        if (state->breakpoints[pos].breakConditionBefore != DWRITE_BREAK_CONDITION_CAN_BREAK)
+        if (has_strong_condition(state->breakpoints[pos].breakConditionBefore, condition))
             return;
         state->breakpoints[pos].breakConditionBefore = condition;
         if (pos > 0)
             state->breakpoints[pos-1].breakConditionAfter = condition;
     }
     else {
-        if (state->breakpoints[pos].breakConditionAfter != DWRITE_BREAK_CONDITION_CAN_BREAK)
+        if (has_strong_condition(state->breakpoints[pos].breakConditionAfter, condition))
             return;
         state->breakpoints[pos].breakConditionAfter = condition;
         if (pos + 1 < state->count)
@@ -403,14 +420,12 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
     state.breakpoints = breakpoints;
     state.count = count;
 
-    /* LB31 - allow breaks everywhere. It will be overridden if needed as
-       other rules dictate. */
     for (i = 0; i < count; i++)
     {
         break_class[i] = get_table_entry(wine_linebreak_table, text[i]);
 
-        breakpoints[i].breakConditionBefore = DWRITE_BREAK_CONDITION_CAN_BREAK;
-        breakpoints[i].breakConditionAfter  = DWRITE_BREAK_CONDITION_CAN_BREAK;
+        breakpoints[i].breakConditionBefore = DWRITE_BREAK_CONDITION_NEUTRAL;
+        breakpoints[i].breakConditionAfter  = DWRITE_BREAK_CONDITION_NEUTRAL;
         breakpoints[i].isWhitespace = !!isspaceW(text[i]);
         breakpoints[i].isSoftHyphen = text[i] == 0x00ad /* Unicode Soft Hyphen */;
         breakpoints[i].padding = 0;
@@ -432,8 +447,10 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
 
     /* LB2 - never break at the start */
     set_break_condition(0, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
-    /* LB3 - always break at the end. This one is ignored. */
+    /* LB3 - always break at the end. */
+    set_break_condition(count - 1, BreakConditionAfter, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
 
+    /* LB4 - LB6 - mandatory breaks. */
     for (i = 0; i < count; i++)
     {
         switch (break_class[i])
@@ -455,16 +472,27 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                 /* LB6 - do not break before hard breaks */
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
+        }
+    }
+
+    /* LB7 - LB8 - explicit breaks and non-breaks */
+    for (i = 0; i < count; i++)
+    {
+        switch (break_class[i])
+        {
             /* LB7 - do not break before spaces */
             case b_SP:
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
             case b_ZW:
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
-            /* LB8 - break before character after zero-width space, skip spaces in-between */
-                while (i < count-1 && break_class[i+1] == b_SP)
-                    i++;
-                set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
+
+                /* LB8 - break before character after zero-width space, skip spaces in-between */
+                j = i;
+                while (j < count-1 && break_class[j+1] == b_SP)
+                    j++;
+                if (j < count-1 && break_class[j+1] != b_ZW)
+                    set_break_condition(j, BreakConditionAfter, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
                 break;
             /* LB8a - do not break between ZWJ and an ideograph, emoji base or emoji modifier */
             case b_ZWJ:
@@ -474,7 +502,7 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
         }
     }
 
-    /* LB9 - LB10 */
+    /* LB9 - LB10 - combining marks */
     for (i = 0; i < count; i++)
     {
         if (break_class[i] == b_CM || break_class[i] == b_ZWJ)
@@ -492,7 +520,10 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                         break_class[i] = b_AL;
                         break;
                     default:
+                    {
                         break_class[i] = break_class[i-1];
+                        set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
+                    }
                 }
             }
             else break_class[i] = b_AL;
@@ -526,22 +557,20 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
             case b_SY:
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
-            /* LB14 */
+            /* LB14 - do not break after OP, even after spaces */
             case b_OP:
-                set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
-                while (i < count-1 && break_class[i+1] == b_SP) {
-                    set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
-                    i++;
-                }
-                break;
-            /* LB15 */
-            case b_QU:
-                j = i+1;
-                while (j < count-1 && break_class[j] == b_SP)
+                j = i;
+                while (j < count-1 && break_class[j+1] == b_SP)
                     j++;
-                if (break_class[j] == b_OP)
-                    for (; j > i; j--)
-                        set_break_condition(j, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
+                set_break_condition(j, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
+                break;
+            /* LB15 - do not break within QU-OP, even with intervening spaces */
+            case b_QU:
+                j = i;
+                while (j < count-1 && break_class[j+1] == b_SP)
+                    j++;
+                if (j < count - 1 && break_class[j+1] == b_OP)
+                    set_break_condition(j, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
             /* LB16 */
             case b_NS:
@@ -552,14 +581,13 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                     for (j++; j <= i; j++)
                         set_break_condition(j, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
-            /* LB17 */
+            /* LB17 - do not break within B2, even with intervening spaces */
             case b_B2:
-                j = i+1;
-                while (j < count && break_class[j] == b_SP)
+                j = i;
+                while (j < count && break_class[j+1] == b_SP)
                     j++;
-                if (break_class[j] == b_B2)
-                    for (; j > i; j--)
-                        set_break_condition(j, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
+                if (j < count - 1 && break_class[j+1] == b_B2)
+                    set_break_condition(j, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
         }
     }
@@ -580,7 +608,8 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
             /* LB20 */
             case b_CB:
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
-                set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
+                if (i < count - 1 && break_class[i+1] != b_QU)
+                    set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
                 break;
             /* LB21 */
             case b_BA:
@@ -589,7 +618,8 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                 set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
             case b_BB:
-                set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
+                if (i < count - 1 && break_class[i+1] != b_CB)
+                    set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
                 break;
             /* LB21a, LB21b */
             case b_HL:
@@ -701,7 +731,7 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                     if (break_class[i+1] == b_IN || break_class[i+1] == b_PO)
                         set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
             }
-            if (break_class[i] == b_PO)
+            if (break_class[i] == b_PR)
             {
                 switch (break_class[i+1])
                 {
@@ -731,7 +761,7 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
                  break_class[i+1] == b_OP)
                 set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
             if (break_class[i] == b_CP &&
-               (break_class[i+1] == b_AL || break_class[i] == b_HL || break_class[i] == b_NU))
+               (break_class[i+1] == b_AL || break_class[i+1] == b_HL || break_class[i+1] == b_NU))
                 set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_MAY_NOT_BREAK, &state);
 
             /* LB30a - break between two RIs if and only if there are an even number of RIs preceding position of the break */
@@ -752,6 +782,13 @@ static HRESULT analyze_linebreaks(const WCHAR *text, UINT32 count, DWRITE_LINE_B
         }
     }
 
+    /* LB31 - allow breaks everywhere else. */
+    for (i = 0; i < count; i++)
+    {
+        set_break_condition(i, BreakConditionBefore, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
+        set_break_condition(i, BreakConditionAfter, DWRITE_BREAK_CONDITION_CAN_BREAK, &state);
+    }
+
     heap_free(break_class);
     return S_OK;
 }
@@ -991,18 +1028,105 @@ static UINT32 get_opentype_language(const WCHAR *locale)
     return language;
 }
 
+static DWRITE_NUMBER_SUBSTITUTION_METHOD get_number_substitutes(IDWriteNumberSubstitution *substitution, WCHAR *digits)
+{
+    struct dwrite_numbersubstitution *numbersubst = unsafe_impl_from_IDWriteNumberSubstitution(substitution);
+    DWRITE_NUMBER_SUBSTITUTION_METHOD method;
+    WCHAR isolang[9];
+    DWORD lctype;
+
+    if (!numbersubst)
+        return DWRITE_NUMBER_SUBSTITUTION_METHOD_NONE;
+
+    lctype = numbersubst->ignore_user_override ? LOCALE_NOUSEROVERRIDE : 0;
+
+    if (numbersubst->method == DWRITE_NUMBER_SUBSTITUTION_METHOD_FROM_CULTURE) {
+        DWORD value;
+
+        method = DWRITE_NUMBER_SUBSTITUTION_METHOD_NONE;
+        if (GetLocaleInfoEx(numbersubst->locale, lctype | LOCALE_IDIGITSUBSTITUTION | LOCALE_RETURN_NUMBER, (WCHAR *)&value, 2)) {
+            switch (value)
+            {
+            case 0:
+                method = DWRITE_NUMBER_SUBSTITUTION_METHOD_CONTEXTUAL;
+                break;
+            case 2:
+                method = DWRITE_NUMBER_SUBSTITUTION_METHOD_NATIONAL;
+                break;
+            case 1:
+            default:
+                if (value != 1)
+                    WARN("Unknown IDIGITSUBSTITUTION value %u, locale %s.\n", value, debugstr_w(numbersubst->locale));
+            }
+        }
+        else
+            WARN("Failed to get IDIGITSUBSTITUTION for locale %s\n", debugstr_w(numbersubst->locale));
+    }
+    else
+        method = numbersubst->method;
+
+    digits[0] = 0;
+    switch (method)
+    {
+    case DWRITE_NUMBER_SUBSTITUTION_METHOD_NATIONAL:
+        GetLocaleInfoEx(numbersubst->locale, lctype | LOCALE_SNATIVEDIGITS, digits, NATIVE_DIGITS_LEN);
+        break;
+    case DWRITE_NUMBER_SUBSTITUTION_METHOD_CONTEXTUAL:
+    case DWRITE_NUMBER_SUBSTITUTION_METHOD_TRADITIONAL:
+        if (GetLocaleInfoEx(numbersubst->locale, LOCALE_SISO639LANGNAME, isolang, sizeof(isolang)/sizeof(isolang[0]))) {
+             static const WCHAR arW[] = {'a','r',0};
+             static const WCHAR arabicW[] = {0x640,0x641,0x642,0x643,0x644,0x645,0x646,0x647,0x648,0x649,0};
+
+             /* For some Arabic locales Latin digits are returned for SNATIVEDIGITS */
+             if (!strcmpW(arW, isolang)) {
+                 strcpyW(digits, arabicW);
+                 break;
+             }
+        }
+        GetLocaleInfoEx(numbersubst->locale, lctype | LOCALE_SNATIVEDIGITS, digits, NATIVE_DIGITS_LEN);
+        break;
+    default:
+        ;
+    }
+
+    if (method != DWRITE_NUMBER_SUBSTITUTION_METHOD_NONE && !*digits) {
+        WARN("Failed to get number substitutes for locale %s, method %d\n", debugstr_w(numbersubst->locale), method);
+        method = DWRITE_NUMBER_SUBSTITUTION_METHOD_NONE;
+    }
+
+    return method;
+}
+
+static void analyzer_dump_user_features(DWRITE_TYPOGRAPHIC_FEATURES const **features,
+        UINT32 const *feature_range_lengths, UINT32 feature_ranges)
+{
+    UINT32 i, j, start;
+
+    if (!TRACE_ON(dwrite) || !features)
+        return;
+
+    for (i = 0, start = 0; i < feature_ranges; i++, start += feature_range_lengths[i]) {
+        TRACE("feature range [%u,%u)\n", start, start + feature_range_lengths[i]);
+        for (j = 0; j < features[i]->featureCount; j++)
+            TRACE("feature %s, parameter %u\n", debugstr_an((char *)&features[i]->features[j].nameTag, 4),
+                    features[i]->features[j].parameter);
+    }
+}
+
 static HRESULT WINAPI dwritetextanalyzer_GetGlyphs(IDWriteTextAnalyzer2 *iface,
     WCHAR const* text, UINT32 length, IDWriteFontFace* fontface, BOOL is_sideways,
     BOOL is_rtl, DWRITE_SCRIPT_ANALYSIS const* analysis, WCHAR const* locale,
     IDWriteNumberSubstitution* substitution, DWRITE_TYPOGRAPHIC_FEATURES const** features,
-    UINT32 const* feature_range_len, UINT32 feature_ranges, UINT32 max_glyph_count,
+    UINT32 const* feature_range_lengths, UINT32 feature_ranges, UINT32 max_glyph_count,
     UINT16* clustermap, DWRITE_SHAPING_TEXT_PROPERTIES* text_props, UINT16* glyph_indices,
     DWRITE_SHAPING_GLYPH_PROPERTIES* glyph_props, UINT32* actual_glyph_count)
 {
     const struct dwritescript_properties *scriptprops;
+    DWRITE_NUMBER_SUBSTITUTION_METHOD method;
     struct scriptshaping_context context;
     struct scriptshaping_cache *cache = NULL;
     BOOL update_cluster, need_vertical;
+    WCHAR digits[NATIVE_DIGITS_LEN];
     IDWriteFontFace1 *fontface1;
     WCHAR *string;
     UINT32 i, g;
@@ -1011,16 +1135,21 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphs(IDWriteTextAnalyzer2 *iface,
 
     TRACE("(%s:%u %p %d %d %s %s %p %p %p %u %u %p %p %p %p %p)\n", debugstr_wn(text, length),
         length, fontface, is_sideways, is_rtl, debugstr_sa_script(analysis->script), debugstr_w(locale), substitution,
-        features, feature_range_len, feature_ranges, max_glyph_count, clustermap, text_props, glyph_indices,
+        features, feature_range_lengths, feature_ranges, max_glyph_count, clustermap, text_props, glyph_indices,
         glyph_props, actual_glyph_count);
 
+    analyzer_dump_user_features(features, feature_range_lengths, feature_ranges);
+
     script = analysis->script > Script_LastId ? Script_Unknown : analysis->script;
 
     if (max_glyph_count < length)
         return E_NOT_SUFFICIENT_BUFFER;
 
-    if (substitution)
-        FIXME("number substitution is not supported.\n");
+    string = heap_alloc(sizeof(WCHAR)*length);
+    if (!string)
+        return E_OUTOFMEMORY;
+
+    method = get_number_substitutes(substitution, digits);
 
     for (i = 0; i < length; i++) {
         /* FIXME: set to better values */
@@ -1035,6 +1164,21 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphs(IDWriteTextAnalyzer2 *iface,
         text_props[i].reserved = 0;
 
         clustermap[i] = i;
+
+        string[i] = text[i];
+        switch (method)
+        {
+        case DWRITE_NUMBER_SUBSTITUTION_METHOD_CONTEXTUAL:
+            if (!is_rtl)
+                break;
+            /* fallthrough */
+        default:
+            if (string[i] >= '0' && string[i] <= '9')
+                string[i] = digits[string[i] - '0'];
+            break;
+        case DWRITE_NUMBER_SUBSTITUTION_METHOD_NONE:
+            ;
+        }
     }
 
     for (; i < max_glyph_count; i++) {
@@ -1045,10 +1189,6 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphs(IDWriteTextAnalyzer2 *iface,
         glyph_props[i].reserved = 0;
     }
 
-    string = heap_alloc(sizeof(WCHAR)*length);
-    if (!string)
-        return E_OUTOFMEMORY;
-
     hr = IDWriteFontFace_QueryInterface(fontface, &IID_IDWriteFontFace1, (void**)&fontface1);
     if (FAILED(hr))
         WARN("failed to get IDWriteFontFace1\n");
@@ -1059,16 +1199,11 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphs(IDWriteTextAnalyzer2 *iface,
         UINT32 codepoint;
 
         if (!update_cluster) {
-            codepoint = decode_surrogate_pair(text, i, length);
-            if (!codepoint) {
-                codepoint = is_rtl ? bidi_get_mirrored_char(text[i]) : text[i];
-                string[i] = codepoint;
-            }
-            else {
-                string[i] = text[i];
-                string[i+1] = text[i+1];
+            codepoint = decode_surrogate_pair(string, i, length);
+            if (!codepoint)
+                codepoint = is_rtl ? bidi_get_mirrored_char(string[i]) : string[i];
+            else
                 update_cluster = TRUE;
-            }
 
             hr = IDWriteFontFace_GetGlyphIndices(fontface, &codepoint, 1, &glyph_indices[g]);
             if (FAILED(hr))
@@ -1136,7 +1271,7 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphPlacements(IDWriteTextAnalyzer2
     UINT32 text_len, UINT16 const* glyphs, DWRITE_SHAPING_GLYPH_PROPERTIES const* glyph_props,
     UINT32 glyph_count, IDWriteFontFace *fontface, FLOAT emSize, BOOL is_sideways, BOOL is_rtl,
     DWRITE_SCRIPT_ANALYSIS const* analysis, WCHAR const* locale, DWRITE_TYPOGRAPHIC_FEATURES const** features,
-    UINT32 const* feature_range_len, UINT32 feature_ranges, FLOAT *advances, DWRITE_GLYPH_OFFSET *offsets)
+    UINT32 const* feature_range_lengths, UINT32 feature_ranges, FLOAT *advances, DWRITE_GLYPH_OFFSET *offsets)
 {
     DWRITE_FONT_METRICS metrics;
     IDWriteFontFace1 *fontface1;
@@ -1145,9 +1280,11 @@ static HRESULT WINAPI dwritetextanalyzer_GetGlyphPlacements(IDWriteTextAnalyzer2
 
     TRACE("(%s %p %p %u %p %p %u %p %.2f %d %d %s %s %p %p %u %p %p)\n", debugstr_wn(text, text_len),
         clustermap, props, text_len, glyphs, glyph_props, glyph_count, fontface, emSize, is_sideways,
-        is_rtl, debugstr_sa_script(analysis->script), debugstr_w(locale), features, feature_range_len,
+        is_rtl, debugstr_sa_script(analysis->script), debugstr_w(locale), features, feature_range_lengths,
         feature_ranges, advances, offsets);
 
+    analyzer_dump_user_features(features, feature_range_lengths, feature_ranges);
+
     if (glyph_count == 0)
         return S_OK;
 
@@ -1197,6 +1334,8 @@ static HRESULT WINAPI dwritetextanalyzer_GetGdiCompatibleGlyphPlacements(IDWrite
         transform, use_gdi_natural, is_sideways, is_rtl, debugstr_sa_script(analysis->script), debugstr_w(locale),
         features, feature_range_lengths, feature_ranges, advances, offsets);
 
+    analyzer_dump_user_features(features, feature_range_lengths, feature_ranges);
+
     if (glyph_count == 0)
         return S_OK;
 
@@ -1493,7 +1632,9 @@ static HRESULT WINAPI dwritetextanalyzer1_GetScriptProperties(IDWriteTextAnalyze
 
 static inline BOOL is_char_from_simple_script(WCHAR c)
 {
-    if (IS_HIGH_SURROGATE(c) || IS_LOW_SURROGATE(c))
+    if (IS_HIGH_SURROGATE(c) || IS_LOW_SURROGATE(c) ||
+            /* LRM, RLM, LRE, RLE, PDF, LRO, RLO */
+            c == 0x200e || c == 0x200f || (c >= 0x202a && c <= 0x202e))
         return FALSE;
     else {
         UINT16 script = get_char_script(c);
@@ -1745,6 +1886,13 @@ static const struct IDWriteNumberSubstitutionVtbl numbersubstitutionvtbl = {
     dwritenumbersubstitution_Release
 };
 
+struct dwrite_numbersubstitution *unsafe_impl_from_IDWriteNumberSubstitution(IDWriteNumberSubstitution *iface)
+{
+    if (!iface || iface->lpVtbl != &numbersubstitutionvtbl)
+        return NULL;
+    return CONTAINING_RECORD(iface, struct dwrite_numbersubstitution, IDWriteNumberSubstitution_iface);
+}
+
 HRESULT create_numbersubstitution(DWRITE_NUMBER_SUBSTITUTION_METHOD method, const WCHAR *locale,
     BOOL ignore_user_override, IDWriteNumberSubstitution **ret)
 {
@@ -1797,14 +1945,14 @@ static ULONG WINAPI fontfallback_AddRef(IDWriteFontFallback *iface)
 {
     struct dwrite_fontfallback *fallback = impl_from_IDWriteFontFallback(iface);
     TRACE("(%p)\n", fallback);
-    return IDWriteFactory4_AddRef(fallback->factory);
+    return IDWriteFactory5_AddRef(fallback->factory);
 }
 
 static ULONG WINAPI fontfallback_Release(IDWriteFontFallback *iface)
 {
     struct dwrite_fontfallback *fallback = impl_from_IDWriteFontFallback(iface);
     TRACE("(%p)\n", fallback);
-    return IDWriteFactory4_Release(fallback->factory);
+    return IDWriteFactory5_Release(fallback->factory);
 }
 
 static int compare_fallback_mapping(const void *a, const void *b)
@@ -1879,28 +2027,41 @@ static HRESULT fallback_map_characters(IDWriteFont *font, const WCHAR *text, UIN
 }
 
 static HRESULT fallback_get_fallback_font(struct dwrite_fontfallback *fallback, const WCHAR *text, UINT32 length,
-    DWRITE_FONT_WEIGHT weight, DWRITE_FONT_STYLE style, DWRITE_FONT_STRETCH stretch, UINT32 *mapped_length, IDWriteFont **mapped_font)
+        DWRITE_FONT_WEIGHT weight, DWRITE_FONT_STYLE style, DWRITE_FONT_STRETCH stretch, UINT32 *mapped_length,
+        IDWriteFont **mapped_font)
 {
     const struct fallback_mapping *mapping;
+    UINT32 i = 0;
     HRESULT hr;
 
+    *mapped_font = NULL;
+
     mapping = find_fallback_mapping(fallback, text[0]);
     if (!mapping) {
-        WARN("no mapping for 0x%x\n", text[0]);
+        WARN("No mapping range for %#x.\n", text[0]);
         return E_FAIL;
     }
 
-    /* now let's see what fallback can handle */
-    hr = create_matching_font((IDWriteFontCollection*)fallback->systemcollection, mapping->family, weight, style, stretch, mapped_font);
-    if (FAILED(hr)) {
-        WARN("failed to create fallback font %s for range [0x%x,0x%x], 0x%08x\n", debugstr_w(mapping->family),
-            mapping->range.first, mapping->range.last, hr);
-        return hr;
+    /* Now let's see what fallback can handle. Pick first font that could be created. */
+    while (mapping->families[i]) {
+        hr = create_matching_font((IDWriteFontCollection *)fallback->systemcollection, mapping->families[i],
+                weight, style, stretch, mapped_font);
+        if (hr == S_OK) {
+            TRACE("Created fallback font for range [%#x, %#x], using family %s.\n", mapping->range.first,
+                    mapping->range.last, debugstr_w(mapping->families[i]));
+            break;
+        }
+        i++;
+    }
+
+    if (!*mapped_font) {
+        WARN("Failed to create fallback font for range [%#x, %#x].\n", mapping->range.first, mapping->range.last);
+        return E_FAIL;
     }
 
     hr = fallback_map_characters(*mapped_font, text, length, mapped_length);
     if (FAILED(hr))
-        WARN("mapping with fallback font %s failed, 0x%08x\n", debugstr_w(mapping->family), hr);
+        WARN("Mapping with fallback family %s failed, hr %#x.\n", debugstr_w(mapping->families[i]), hr);
 
     if (!*mapped_length) {
         IDWriteFont_Release(*mapped_font);
@@ -1980,7 +2141,7 @@ static const IDWriteFontFallbackVtbl fontfallbackvtbl = {
     fontfallback_MapCharacters
 };
 
-HRESULT create_system_fontfallback(IDWriteFactory4 *factory, IDWriteFontFallback **ret)
+HRESULT create_system_fontfallback(IDWriteFactory5 *factory, IDWriteFontFallback **ret)
 {
     struct dwrite_fontfallback *fallback;
 
@@ -1994,7 +2155,7 @@ HRESULT create_system_fontfallback(IDWriteFactory4 *factory, IDWriteFontFallback
     fallback->factory = factory;
     fallback->mappings = fontfallback_neutral_data;
     fallback->count = sizeof(fontfallback_neutral_data)/sizeof(fontfallback_neutral_data[0]);
-    IDWriteFactory4_GetSystemFontCollection(fallback->factory, FALSE, &fallback->systemcollection, FALSE);
+    IDWriteFactory5_GetSystemFontCollection(fallback->factory, FALSE, &fallback->systemcollection, FALSE);
 
     *ret = &fallback->IDWriteFontFallback_iface;
     return S_OK;
