@@ -825,13 +825,13 @@ static void HTMLBodyElement_unlink(HTMLDOMNode *iface)
     }
 }
 
-static event_target_t **HTMLBodyElement_get_event_target_ptr(HTMLDOMNode *iface)
+static EventTarget *HTMLBodyElement_get_event_target(HTMLDOMNode *iface)
 {
     HTMLBodyElement *This = impl_from_HTMLDOMNode(iface);
 
-    return This->textcont.element.node.doc
-        ? &This->textcont.element.node.doc->body_event_target
-        : &This->textcont.element.node.event_target.ptr;
+    return This->textcont.element.node.doc && This->textcont.element.node.doc->window
+        ? &This->textcont.element.node.doc->window->event_target
+        : &This->textcont.element.node.event_target;
 }
 
 static BOOL HTMLBodyElement_is_text_edit(HTMLDOMNode *iface)
@@ -857,13 +857,14 @@ static const cpc_entry_t HTMLBodyElement_cpc[] = {
 };
 
 static const NodeImplVtbl HTMLBodyElementImplVtbl = {
+    &CLSID_HTMLBody,
     HTMLBodyElement_QI,
     HTMLElement_destructor,
     HTMLBodyElement_cpc,
     HTMLElement_clone,
     HTMLElement_handle_event,
     HTMLElement_get_attr_col,
-    HTMLBodyElement_get_event_target_ptr,
+    HTMLBodyElement_get_event_target,
     NULL,
     NULL,
     NULL,
